//************************************************************
//  University of Patras
//	Department of Electrical Engineering & Computer Technology
//
//	Course: Parallel Programming in AI
//	Week 04: OpenMp
//
//  Description: Collatz in openMP
//  	omp05Collatz.c: the hello work openmp version
//	Author: Evangelos Dermatas
//	v0.1
/*
*********************************************************************
gcc -v:	gcc version 7.4.0 (Ubuntu 7.4.0-1ubuntu1~18.04.1)
OS: 	Linux Ubuntu 18.04
CPU:	8-core, Intel(R) Core(TM) i7-2630QM CPU @ 2.00GHz, 3991 BogoMips
*********************************************************************
Compiling:	gcc omp05Collatz.c -o omp05Collatz -fopenmp
Executing:	./omp05Collatz

Output in my Laptop:

*/
//************************************************************


// *******************************************************************
#pragma GCC optimize("O3","unroll-loops","omit-frame-pointer","inline") 
#pragma GCC option("arch=native","tune=native","no-zero-upper") 
//************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <omp.h>


// Maximum number of steps
#define NSTEPS  100000

// Maximum natural number
#define N       10000000

unsigned long int Hist[NSTEPS+1] ;

// ***************************************
void CollatzSerial(void)
{
    int q, k ;

//#1 pragma omp parallel for
//#2 pragma omp parallel for private(k,q)
//#3 pragma omp parallel for private(k,q)  schedule(dynamic,1000)
//#4 pragma omp parallel for private(k,q)  static(runtime,1000)
#pragma omp parallel for private(k,q)  schedule(runtime,1000)
    for( int i = 1 ; i < N ; i++  ) {
        for ( k = 0, q = i ; k<NSTEPS ; k++ ) {
            if ( q == 1 ) {
#pragma omp atomic
                Hist[k]++ ;
                break ;
            }
            if ( q%2 == 0 )
                q /= 2 ;
            else
                q=3*q+1;
        }
        if ( k == NSTEPS )
#pragma omp atomic
            Hist[NSTEPS]++ ;
    }
}

// ***************************************
void PrintResults()
{
    unsigned long int max=0 ;
    int ind=0;

for ( int k = 0 ; k < NSTEPS ; k++ )
    if ( max < Hist[k] ) {
        ind = k ;
        max = Hist[k] ;
    }
    printf( "Steps %d, Frequency %lu\n", ind, max) ;
}


// ***************************************
int main()
{
//    omp_set_num_threads(8);
    CollatzSerial() ;
    PrintResults() ;
    return 0 ;
}

